/*
* boxes.cpp
* An exercise in positioning simple boxes using projection/modelview
* matrices and standard transforms.
*
* Adapted for WebGL by Alex Clarke, 2016.
* Adapted for WebGL2 by Alex Clarke, 2020.
* Shape class and various fixes by Alex Clarke, 2024
*/




//----------------------------------------------------------------------------
// State Variable Setup 
//----------------------------------------------------------------------------

// This variable will store the WebGL rendering context
let gl;

//Variables for Transformation Matrices
let mv = new mat4();
let p = new mat4();
let mvLoc, projLoc;

//Model state variables
let shoulder = 0, elbow = 0;


//----------------------------------------------------------------------------
// A Simple Drawable Shape Manager Class
//----------------------------------------------------------------------------
class Shape {
   /**
    *  Construct a shape
    */
   constructor() {
      this.type = null;
      this.start = 0;
      this.length = 0;
      this.points = [];
      this.colours = [];
   }

   /** 
    * Adds shape to global points and colour arrays  
    * and calculates draw start index and length
    * 
    * @param type - primitive to draw data with. GL context must exist to refer to this.
    */
   load(type) {
      if (this.points.length != this.colours.length) {
         throw new Error("positions array and colours array are different lengths");
      }
      this.type = type;
      this.start = points.length;
      this.length = this.points.length;
      points = points.concat(this.points);
      colours = colours.concat(this.colours);
   }

   /**
    * Encapsulates code to draw the shape
    */
   draw() {
      gl.drawArrays(this.type, this.start, this.length);
   }
}


// Define a tidy collection of shapes to use
let shapes = {
   wireCube: new Shape(),
   solidCube: new Shape(),
   axes: new Shape(),
};


//----------------------------------------------------------------------------
// Define Shape Data 
//----------------------------------------------------------------------------

//Some colours to use
let red = vec4(1.0, 0.0, 0.0, 1.0);
let green = vec4(0.0, 1.0, 0.0, 1.0);
let blue = vec4(0.0, 0.0, 1.0, 1.0);
let lightRed = vec4(1.0, 0.5, 0.5, 1.0);
let lightGreen = vec4(0.5, 1.0, 0.5, 1.0);
let lightBlue = vec4(0.5, 0.5, 1.0, 1.0);
let white = vec4(1.0, 1.0, 1.0, 1.0);

//Generate Axis Data: use LINES to draw. Three axes in red, green and blue
shapes.axes.points =
   [
      vec4(2.0, 0.0, 0.0, 1.0), //x axis, will be green
      vec4(-2.0, 0.0, 0.0, 1.0),
      vec4(0.0, 2.0, 0.0, 1.0), //y axis, will be red
      vec4(0.0, -2.0, 0.0, 1.0),
      vec4(0.0, 0.0, 2.0, 1.0), //z axis, will be blue
      vec4(0.0, 0.0, -2.0, 1.0)
   ];

shapes.axes.colours =
   [
      green, green,
      red, red,
      blue, blue
   ];


//Define points for a unit cube
let cubeVerts = [
   vec4(0.5, 0.5, 0.5, 1), //0
   vec4(0.5, 0.5, -0.5, 1), //1
   vec4(0.5, -0.5, 0.5, 1), //2
   vec4(0.5, -0.5, -0.5, 1), //3
   vec4(-0.5, 0.5, 0.5, 1), //4
   vec4(-0.5, 0.5, -0.5, 1), //5
   vec4(-0.5, -0.5, 0.5, 1), //6
   vec4(-0.5, -0.5, -0.5, 1), //7
];

//Look up patterns from cubeVerts for different primitive types
//Wire Cube - draw with LINE_STRIP
let wireCubeLookups = [
   0, 4, 6, 2, 0, //front
   1, 0, 2, 3, 1, //right
   5, 1, 3, 7, 5, //back
   4, 5, 7, 6, 4, //right
   4, 0, 1, 5, 4, //top
   6, 7, 3, 2, 6, //bottom
];

//Solid Cube - draw with TRIANGLES, 2 triangles per face
let solidCubeLookups = [
   0, 4, 6, 0, 6, 2, //front
   1, 0, 2, 1, 2, 3, //right
   5, 1, 3, 5, 3, 7,//back
   4, 5, 7, 4, 7, 6,//left
   4, 0, 1, 4, 1, 5,//top
   6, 7, 3, 6, 3, 2,//bottom
];

//Expand Wire Cube data: this wire cube will be white...
for (let i = 0; i < wireCubeLookups.length; i++) {
   shapes.wireCube.points.push(cubeVerts[wireCubeLookups[i]]);
   shapes.wireCube.colours.push(white);
}

//Expand Solid Cube data: each face will be a different color so you can see
//    the 3D shape better without lighting.
let colorNum = 0;
let colorList = [lightBlue, lightGreen, lightRed, blue, red, green];
for (let i = 0; i < solidCubeLookups.length; i++) {
   shapes.solidCube.points.push(cubeVerts[solidCubeLookups[i]]);
   shapes.solidCube.colours.push(colorList[colorNum]);
   if (i % 6 == 5) colorNum++; //Switch color for every face. 6 vertices/face
}

//load data into points and colours arrays - runs once as page loads.
let points = [];
let colours = [];


//----------------------------------------------------------------------------
// Initialization Event Function
//----------------------------------------------------------------------------

window.onload = function init() {
   // Set up a WebGL Rendering Context in an HTML5 Canvas
   let canvas = document.getElementById("gl-canvas");
   gl = canvas.getContext("webgl2");
   if (!gl) {
      canvas.parentNode.innerHTML("Cannot get WebGL2 Rendering Context");
   }

   //  Configure WebGL
   //  eg. - set a clear color
   //      - turn on depth testing
   gl.viewport(0, 0, gl.drawingBufferWidth, gl.drawingBufferHeight);
   gl.enable(gl.DEPTH_TEST);
   gl.clearColor(0.0, 0.0, 0.0, 1.0);
   gl.enable(gl.CULL_FACE);

   //  Load shaders and initialize attribute buffers
   let program = initShaders(gl, "shader.vert", "shader.frag");
   gl.useProgram(program);

   // Set up data to draw
   // Mostly done globally in this program all we are doing here is adding
   // the data to the points and colour arrays
   shapes.wireCube.load(gl.LINE_STRIP);
   shapes.solidCube.load(gl.TRIANGLES);
   shapes.axes.load(gl.LINES);

   // Load the data into GPU data buffers and
   // Associate shader attributes with corresponding data buffers
   //***Vertices***
   let vertexBuffer = gl.createBuffer();
   gl.bindBuffer(gl.ARRAY_BUFFER, vertexBuffer);
   gl.bufferData(gl.ARRAY_BUFFER, flatten(points), gl.STATIC_DRAW);
   program.vPosition = gl.getAttribLocation(program, "vPosition");
   gl.vertexAttribPointer(program.vPosition, 4, gl.FLOAT, gl.FALSE, 0, 0);
   gl.enableVertexAttribArray(program.vPosition);

   //***Colours***
   let colorBuffer = gl.createBuffer();
   gl.bindBuffer(gl.ARRAY_BUFFER, colorBuffer);
   gl.bufferData(gl.ARRAY_BUFFER, flatten(colours), gl.STATIC_DRAW);
   program.vColor = gl.getAttribLocation(program, "vColor");
   gl.vertexAttribPointer(program.vColor, 4, gl.FLOAT, gl.FALSE, 0, 0);
   gl.enableVertexAttribArray(program.vColor);

   // Get addresses of shader uniforms
   projLoc = gl.getUniformLocation(program, "p");
   mvLoc = gl.getUniformLocation(program, "mv");

   //Set up viewport - see WebGL Anti-Patterns link
   //gl.viewport(0, 0, gl.drawingBufferWidth, gl.drawingBufferHeight);

   //Set up projection matrix
   p = perspective(45.0, 1.0, 0.1, 100.0);
   gl.uniformMatrix4fv(projLoc, gl.FALSE, flatten(transpose(p)));

   requestAnimationFrame(render);
};



//----------------------------------------------------------------------------
// Rendering Event Function
//----------------------------------------------------------------------------
function render() {

   gl.clear(gl.DEPTH_BUFFER_BIT | gl.COLOR_BUFFER_BIT);

   //Set initial view
   let eye = vec3(0.0, 0.0, 10.0);
   let at = vec3(0.0, 0.0, 0.0);
   let up = vec3(0.0, 1.0, 0.0);

   mv = lookAt(eye, at, up);


   gl.uniformMatrix4fv(mvLoc, gl.FALSE, flatten(transpose(mv)));
   shapes.axes.draw();


}