/*
* Copyright (c) 1993-1997, Silicon Graphics, Inc.
* ALL RIGHTS RESERVED 
* Permission to use, copy, modify, and distribute this software for 
* any purpose and without fee is hereby granted, provided that the above
* copyright notice appear in all copies and that both the copyright notice
* and this permission notice appear in supporting documentation, and that 
* the name of Silicon Graphics, Inc. not be used in advertising
* or publicity pertaining to distribution of the software without specific,
* written prior permission. 
*
* THE MATERIAL EMBODIED ON THIS SOFTWARE IS PROVIDED TO YOU "AS-IS"
* AND WITHOUT WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR OTHERWISE,
* INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF MERCHANTABILITY OR
* FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL SILICON
* GRAPHICS, INC.  BE LIABLE TO YOU OR ANYONE ELSE FOR ANY DIRECT,
* SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY
* KIND, OR ANY DAMAGES WHATSOEVER, INCLUDING WITHOUT LIMITATION,
* LOSS OF PROFIT, LOSS OF USE, SAVINGS OR REVENUE, OR THE CLAIMS OF
* THIRD PARTIES, WHETHER OR NOT SILICON GRAPHICS, INC.  HAS BEEN
* ADVISED OF THE POSSIBILITY OF SUCH LOSS, HOWEVER CAUSED AND ON
* ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE
* POSSESSION, USE OR PERFORMANCE OF THIS SOFTWARE.
* 
* US Government Users Restricted Rights 
* Use, duplication, or disclosure by the Government is subject to
* restrictions set forth in FAR 52.227.19(c)(2) or subparagraph
* (c)(1)(ii) of the Rights in Technical Data and Computer Software
* clause at DFARS 252.227-7013 and/or in similar or successor
* clauses in the FAR or the DOD or NASA FAR Supplement.
* Unpublished-- rights reserved under the copyright laws of the
* United States.  Contractor/manufacturer is Silicon Graphics,
* Inc., 2011 N.  Shoreline Blvd., Mountain View, CA 94039-7311.
*
* OpenGL(R) is a registered trademark of Silicon Graphics, Inc.
*/

/*
* robot.cpp
* This program shows how to composite modeling transformations
* to draw translated and rotated hierarchical models.
* Interaction:  pressing the s and e keys (shoulder and elbow)
* alters the rotation of the robot arm.
*
* Adapted for OpenGL 3.2 Core Profile by Alex Clarke, 2012, with
* libraries provided with Interactive Computer Graphics 6th Ed. 
* by Dr. Edward Angel and Dave Shreiner
*
* Adapted from the above for WebGL by Alex Clarke, 2016, with
* libraries provided with Interactive Computer Graphics 7th Ed.
* by Dr. Edward Angel and Dave Shreiner
*
* Adapted for WebGL2 by Alex Clarke, 2019.
* Shape class and various fixes by Alex Clarke, 2024
*/


//----------------------------------------------------------------------------
// State Variable Setup 
//----------------------------------------------------------------------------

// This variable will store the WebGL rendering context
let gl;

//Variables for Transformation Matrices
let mv = new mat4();
let p = new mat4();
let mvLoc, projLoc;

//Model state variables
let shoulder = 0, elbow = 0;


//----------------------------------------------------------------------------
// A Simple Drawable Shape Manager Class
//----------------------------------------------------------------------------
class Shape {
   /**
    *  Construct a shape
    */
   constructor() {
      this.type = null;
      this.start = 0;
      this.length = 0;
      this.points = [];
      this.colours = [];
   }

   /** 
    * Adds shape to global points and colour arrays  
    * and calculates draw start index and length
    * 
    * @param type - primitive to draw data with. GL context must exist to refer to this.
    */
   load(type) {
      if (this.points.length != this.colours.length) {
         throw new Error("positions array and colours array are different lengths");
      }
      this.type = type;
      this.start = points.length;
      this.length = this.points.length;
      points = points.concat(this.points);
      colours = colours.concat(this.colours);
   }

   /**
    * Encapsulates code to draw the shape
    */
   draw() {
      gl.drawArrays(this.type, this.start, this.length);
   }
}


// Define a tidy collection of shapes to use
let shapes = {
   wireCube: new Shape(),
   solidCube: new Shape(),
   axes: new Shape(),
};


//----------------------------------------------------------------------------
// Define Shape Data 
//----------------------------------------------------------------------------

//Some colours to use
let red = vec4(1.0, 0.0, 0.0, 1.0);
let green = vec4(0.0, 1.0, 0.0, 1.0);
let blue = vec4(0.0, 0.0, 1.0, 1.0);
let lightRed = vec4(1.0, 0.5, 0.5, 1.0);
let lightGreen = vec4(0.5, 1.0, 0.5, 1.0);
let lightBlue = vec4(0.5, 0.5, 1.0, 1.0);
let white = vec4(1.0, 1.0, 1.0, 1.0);

//Generate Axis Data: use LINES to draw. Three axes in red, green and blue
shapes.axes.points =
   [
      vec4(2.0, 0.0, 0.0, 1.0), //x axis, will be green
      vec4(-2.0, 0.0, 0.0, 1.0),
      vec4(0.0, 2.0, 0.0, 1.0), //y axis, will be red
      vec4(0.0, -2.0, 0.0, 1.0),
      vec4(0.0, 0.0, 2.0, 1.0), //z axis, will be blue
      vec4(0.0, 0.0, -2.0, 1.0)
   ];

shapes.axes.colours =
   [
      green, green,
      red, red,
      blue, blue
   ];


//Define points for a unit cube
let cubeVerts = [
   vec4(0.5, 0.5, 0.5, 1), //0
   vec4(0.5, 0.5, -0.5, 1), //1
   vec4(0.5, -0.5, 0.5, 1), //2
   vec4(0.5, -0.5, -0.5, 1), //3
   vec4(-0.5, 0.5, 0.5, 1), //4
   vec4(-0.5, 0.5, -0.5, 1), //5
   vec4(-0.5, -0.5, 0.5, 1), //6
   vec4(-0.5, -0.5, -0.5, 1), //7
];

//Look up patterns from cubeVerts for different primitive types
//Wire Cube - draw with LINE_STRIP
let wireCubeLookups = [
   0, 4, 6, 2, 0, //front
   1, 0, 2, 3, 1, //right
   5, 1, 3, 7, 5, //back
   4, 5, 7, 6, 4, //right
   4, 0, 1, 5, 4, //top
   6, 7, 3, 2, 6, //bottom
];

//Solid Cube - draw with TRIANGLES, 2 triangles per face
let solidCubeLookups = [
   0, 4, 6, 0, 6, 2, //front
   1, 0, 2, 1, 2, 3, //right
   5, 1, 3, 5, 3, 7,//back
   4, 5, 7, 4, 7, 6,//left
   4, 0, 1, 4, 1, 5,//top
   6, 7, 3, 6, 3, 2,//bottom
];

//Expand Wire Cube data: this wire cube will be white...
for (let i = 0; i < wireCubeLookups.length; i++) {
   shapes.wireCube.points.push(cubeVerts[wireCubeLookups[i]]);
   shapes.wireCube.colours.push(white);
}

//Expand Solid Cube data: each face will be a different colour so you can see
//    the 3D shape better without lighting.
let colourNum = 0;
let colourList = [lightBlue, lightGreen, lightRed, blue, red, green];
for (let i = 0; i < solidCubeLookups.length; i++) {
   shapes.solidCube.points.push(cubeVerts[solidCubeLookups[i]]);
   shapes.solidCube.colours.push(colourList[colourNum]);
   if (i % 6 == 5) colourNum++; //Switch colour for every face. 6 vertices/face
}

//load data into points and colours arrays - runs once as page loads.
let points = [];
let colours = [];

//----------------------------------------------------------------------------
// Initialization Event Function
//----------------------------------------------------------------------------

window.onload = function init() {
   // Set up a WebGL Rendering Context in an HTML5 Canvas
   var canvas = document.getElementById("gl-canvas");
   gl = canvas.getContext("webgl2");
   if (!gl) {
      canvas.parentNode.innerHTML("Cannot get WebGL2 Rendering Context");
   }

   //  Configure WebGL
   //  eg. - set a clear colour
   //      - turn on depth testing
   gl.viewport(0, 0, gl.drawingBufferWidth, gl.drawingBufferHeight);
   gl.enable(gl.DEPTH_TEST);
   gl.clearColor(0.0, 0.0, 0.0, 1.0);
   gl.enable(gl.CULL_FACE);

   //  Load shaders and initialize attribute buffers
   var program = initShaders(gl, "shader.vert", "shader.frag");
   gl.useProgram(program);

   // Set up data to draw
   // Mostly done globally in this program all we are doing here is adding
   // the data to the points and colour arrays
   shapes.wireCube.load(gl.LINE_STRIP);
   shapes.solidCube.load(gl.TRIANGLES);
   shapes.axes.load(gl.LINES);


   // Load the data into GPU data buffers and
   // Associate shader attributes with corresponding data buffers
   //***Vertices***
   var vertexBuffer = gl.createBuffer();
   gl.bindBuffer(gl.ARRAY_BUFFER, vertexBuffer);
   gl.bufferData(gl.ARRAY_BUFFER, flatten(points), gl.STATIC_DRAW);
   program.vPosition = gl.getAttribLocation(program, "vPosition");
   gl.vertexAttribPointer(program.vPosition, 4, gl.FLOAT, gl.FALSE, 0, 0);
   gl.enableVertexAttribArray(program.vPosition);

   //***colours***
   var colourBuffer = gl.createBuffer();
   gl.bindBuffer(gl.ARRAY_BUFFER, colourBuffer);
   gl.bufferData(gl.ARRAY_BUFFER, flatten(colours), gl.STATIC_DRAW);
   program.vColor = gl.getAttribLocation(program, "vColor");
   gl.vertexAttribPointer(program.vColor, 4, gl.FLOAT, gl.FALSE, 0, 0);
   gl.enableVertexAttribArray(program.vColor);

   // Get addresses of shader uniforms
   projLoc = gl.getUniformLocation(program, "p");
   mvLoc = gl.getUniformLocation(program, "mv");

   //Set up projection matrix
   var aspect = canvas.clientWidth / canvas.clientHeight;
   //p = ortho(-3.4*aspect, 3.4*aspect, -3.4, 3.4, 1.0, 20.0);
   p = perspective(40.0, aspect, 0.1, 100.0);

   gl.uniformMatrix4fv(projLoc, gl.FALSE, flatten(transpose(p)));

   //Set initial view
   var eye = vec3(0.0, 1.0, 10.0);
   var at = vec3(0.0, 0.0, 0.0);
   var up = vec3(0.0, 1.0, 0.0);

   mv = lookAt(eye, at, up);

   //Animate - draw continuously
   requestAnimationFrame(animate);
};



//----------------------------------------------------------------------------
// Animation and Rendering Event Functions
//----------------------------------------------------------------------------

//animate()
//updates and displays the model based on elapsed time
//sets up another animation event as soon as possible
var prevTime = 0;
function animate() {
   requestAnimationFrame(animate);

   //Do time corrected animation
   var curTime = new Date().getTime();
   if (prevTime != 0) {
      //Calculate elapsed time in seconds
      var timePassed = (curTime - prevTime) / 1000.0;
      //Update any active animations 
      handleKeys(timePassed);
   }
   prevTime = curTime;

   //Draw
   render();
}

function render() {
   gl.clear(gl.DEPTH_BUFFER_BIT | gl.COLOR_BUFFER_BIT);

   var armShape = shapes.solidCube;
   var matStack = [];

   //Save view transform
   matStack.push(mv);
   {

      //Position Shoulder Joint
      mv = mult(mv, translate(-2.0, 0.0, 0.0));
      //Shoulder Joint
      mv = mult(mv, rotate(shoulder, vec3(0, 0, 1)));
      //Position Upper Arm Cube
      mv = mult(mv, translate(1.0, 0.0, 0.0));
      //Scale and Draw Upper Arm
      matStack.push(mv);
      {
         mv = mult(mv, scale(2.0, 0.4, 1.0));
         gl.uniformMatrix4fv(mvLoc, gl.FALSE, flatten(transpose(mv)));
         armShape.draw();
      }
      //Undo Scale
      mv = matStack.pop();


      //Position Elbow Joint
      mv = mult(mv, translate(1.0, 0.0, 0.0));
      //Elbow Joint
      mv = mult(mv, rotate(elbow, vec3(0, 0, 1)));
      //Position Forearm Cube
      mv = mult(mv, translate(1, 0.0, 0.0));
      //Scale and Draw Forearm
      matStack.push(mv);
      {
         mv = mult(mv, scale(2.0, 0.4, 1.0));
         gl.uniformMatrix4fv(mvLoc, gl.FALSE, flatten(transpose(mv)));
         armShape.draw();
      }
      //Undo Scale
      mv = matStack.pop();

   }
   //Restore mv to initial state
   mv = matStack.pop();

}



//----------------------------------------------------------------------------
// Keyboard Event Functions
//----------------------------------------------------------------------------

//This array will hold the pressed or unpressed state of every key
var currentlyPressedKeys = [];

//Store current state of shift key
var shift;

document.onkeydown = function handleKeyDown(event) {
   currentlyPressedKeys[event.keyCode] = true;
   shift = event.shiftKey;

   //Get unshifted key character
   var c = event.keyCode;
   var key = String.fromCharCode(c);

   //Place key down detection code here
}

document.onkeyup = function handleKeyUp(event) {
   currentlyPressedKeys[event.keyCode] = false;
   shift = event.shiftKey;

   //Get unshifted key character
   var c = event.keyCode;
   var key = String.fromCharCode(c);

   //Place key up detection code here
}

//isPressed(c)
//Utility function to lookup whether a key is pressed
//Only works with unshifted key symbol
// ie: use "E" not "e"
//     use "5" not "%"
function isPressed(c) {
   var code = c.charCodeAt(0);
   return currentlyPressedKeys[code];
}

//handleKeys(timePassed)
//Continuously called from animate to cause model updates based on
//any keys currently being held down
function handleKeys(timePassed) {
   //Place continuous key actions here - anything that should continue while a key is
   //held

   //Calculate how much to move based on time since last update
   var s = 90.0; //rotation speed in degrees per second
   var d = s * timePassed; //degrees to rotate on this frame

   //Shoulder Updates
   if (shift && isPressed("S")) {

      if (shoulder < 90) shoulder = (shoulder + d);
      else shoulder = 90;
   }
   if (!shift && isPressed("S")) {
      if (shoulder > -90) shoulder = (shoulder - d);
      else shoulder = -90;
   }

   //Elbow Updates
   if (shift && isPressed("E")) {
      if (elbow < 0) elbow = (elbow + d);
      else elbow = 0;
   }
   if (!shift && isPressed("E")) {
      if (elbow > -144) elbow = (elbow - d);
      else elbow = -144;
   }
}
